from .printer import Printer
from .table import Table
from .fill import Fill
from .flex import Flex
from .card import Card
from . import context
from .hr import HR
from .line import Line
from .text import Text
from .pretty import pretty
print = Printer()



import io

def oid(obj):
    return 'None' if obj==None else str(hex(id(obj)))[-4:]

def ostr(obj):
    return f'<{obj.__class__.__name__}#{oid(obj)}> '


def print_ctx_trace(obj, to_parent=None):
    o = obj
    while o != None and o != to_parent:
        line = ('' if obj == o else '  -> ') + ostr(o) + ' '
        for trace in o.ctx_trace():
            line += ', '.join(f'{k.names[0]}={v}' for k,v in trace.items()) + ' -- '
        print(line)
        o = o.parent




def context_info(obj, parent=None):
    def _s(o):
        if o == None or o == parent: return ''
        style = ' '.join(f'{k}:{repr(v)}' for k,v in o.ctx_local.items()) if hasattr(o, 'ctx_local') else ''
        if hasattr(o, 'parent'):
            return f"{style} {'' if o.parent==None else ostr(o.parent)}  {_s(o.parent)}"
        return f'{style}'
    return _s(obj)



def ostr_ctx(obj, parent=None):
    return f'{ostr(obj)} {context_info(obj, parent)}'


def debug_dump(obj, parent=None, depth=0):
    s = ['  '*depth + ostr_ctx(obj, parent) + f' ---{obj}---']
    for sub in obj.each_child() if hasattr(obj,'each_child') else []:
        s += debug_dump(sub, obj, depth + 1)
    return s



def flat(widget, **kwargs):
    return [r.styled()[0] for r in widget.flatten(**kwargs)]


def styled(widget, **kwargs):
    return [r.styled() for r in widget.flatten(**kwargs)]


def printer(**kwargs):
    from print_ext.printer import Printer
    o = io.StringIO()
    p = Printer(stream=o, **kwargs)
    return o,p
   



import pytest
from print_ext.card import Card
from print_ext.line import SMark as SM
from .testutil import flat, printer, styled
from print_ext.flex import Flex


def test_card_hello():
    c = Card('hello\t', 'to the great big\vworld')
    assert(flat(c) == [
        '┌────┤ hello ├─────┐',
        '│ to the great big │',
        '│ world            │',
        '└──────────────────┘',])


def test_card_multi_title():
    assert(flat(Card('a\vb\vc\tinner body\t')) == [
        '    │ a │     ',
        '    │ b │     ',
        '┌───┤ c ├────┐',
        '│ inner body │',
        '└────────────┘',   
    ])


def test_card_empty():
    assert(flat(Card()) == ['┌──┐', '└──┘'])


def test_card_notitle():
    assert(flat(Card('\tHello to the\vworld')) == [
        '┌──────────────┐', 
        '│ Hello to the │', 
        '│ world        │', 
        '└──────────────┘'])



def test_card_clone():
    c = Card('title\tbody')
    f = Flex(c, '\t', c)
    assert(flat(f) == ['┌┤ title ├┐┌┤ title ├┐', '│ body    ││ body    │', '└─────────┘└─────────┘'])



def test_border_hilite_bug():
    ''' Show a portion of the table while building.  The col widths may change, so use fixed col widths
    '''
    o,p = printer(color=True)
    p.card('\berr Error:\t', 'Some problem')
    c = Card('\berr Error:\t', 'Some problem')
    assert(styled(c) == [
        ('┌──┤ Error: ├──┐', [SM('dem',0,5), SM('err',5,11), SM('dem',11,16)]),
        ('│ Some problem │', [SM('dem',0,2), SM('dem',14,16)]),
        ('└──────────────┘', [SM('dem',0,16)]),
    ])
    print(o.getvalue())
    print(repr(o.getvalue()))
    assert(o.getvalue() == '\x1b[2;37m┌──┤ \x1b[1;31mError:\x1b[0m ├──┐\n\x1b[2;37m│ \x1b[0mSome problem\x1b[2;37m │\x1b[0m\n\x1b[2;37m└──────────────┘\x1b[0m\n')




from functools import reduce
from .context import Context, CVar
from .cache import cache
from .widget import Widget


class ProxyContext(Context):
    def __init__(self, eoc=True, **kwargs):
        self.eoc = eoc
        super().__init__(**kwargs)

    def is_empty(self):
        return not self.ctx_local and (not self.parent or self.parent.is_empty())

    def ctx_flatten(self):
        return {} if self.is_empty() else super().ctx_flatten()

        

class LangCVar(CVar):
    def canon(self, val):
        if isinstance(val, str):
            return tuple((val.lower()+'___').split('_')[:3])
        return val

Context.define(LangCVar('lang'))



class Rich(Widget):
    ''' This is a mixin for adding rich-text parsing capabilities via the __call__ method.

    :Special characters:
        \\b
            Put an attribute on the stack.
            `\\b<key>=<value><auto-pop>`

            `\\b=R`   This gets auto-popped at the end of the call
            `\\b=y$`  This gets auto-popped at the end of the string
            `\\b=r!`  This doesn't get auto-popped.  Pop it manually with 
            `\\b=`    Pop an attribute.  It is ok to pop too many times.

        \\v
            If a string contains \\n then it is first processed with cleandoc() which removes
            preceeding whitespace and common whitespace on each line.  \\v dosen't trigger
            this logic and gets replaced as \\n

        \\t
            Triggers a move to the next 'cell'.  The meaning of 'cell' is defined by the baseclass.

        \\f
            Separate language versions
            \\f<lang>_<region>_<dialect>

            Region and dialect are optional.  The best language match will be selected.
            
    :Usage:
        Subclass this and then implement the following methods as a state machine.
            rich_append(el)
                el will be flattenable
            rich_newline()
                This is optional.  By default throws.
            rich_tab()
                This is optional.  By default throws.
    '''

    def __init__(self, *args, **kwargs):
        self.rich_stream = []
        super().__init__(**kwargs)
        self(*args)


    def __call__(self, *els, **kwargs):
        return self.append(*els, **kwargs)


    def append(self, *els, **kwargs):
        ''' Add any kind of element to this.

        If el is a str then it is parsed as rich-text. (b,f,t,v escape codes)
        If it is a Context object, then it is added as is.
        Otherwise, turn it into a Line() and add that.
        '''
        self.ctx_cur = ProxyContext(**kwargs)
        for el in els:
            if isinstance(el, str):
                lines = _unindent(el).replace('\v','\n').split('\n') # FIXME: Removed self.__l10n(el)
                self.__append_line(lines[0])
                for line in lines[1:]:
                    self.rich_newline(self.ctx_cur.ctx_flatten())
                    self.__append_line(line)
                while self.__pop(False): pass # Pop end-of-string
            elif isinstance(el, Context):
                self.rich_append(el, self.ctx_cur.ctx_flatten())
            else:
                self.rich_append(str(el), self.ctx_cur.ctx_flatten())
        self.changed_size()
        return self


    def __append_line(self, line):
        tabs = line.split('\t')
        self.__append_tab(tabs[0])
        for tab in tabs[1:]:
            self.rich_tab(self.ctx_cur.ctx_flatten())
            self.__append_tab(tab)
    

    def __append_tab(self, tab):
        parts = tab.split('\b')
        if parts[0]: self.rich_append(parts[0], self.ctx_cur.ctx_flatten())
        for part in parts[1:]:
            part = part.split(' ',1)
            eoc = part[0].endswith('$')
            part, txt = part[0][:-1] if eoc else part[0], (part[1] if len(part) > 1 else '')
            if part:
                part = part.split('=')
                kwargs = {'':part[0]} if len(part) == 1 else {part[0]:part[1]}
                self.ctx_cur = ProxyContext(eoc=eoc, parent=self.ctx_cur, **kwargs)
            else: # empty \b
                self.__pop(eoc)
            if txt: self.rich_append(txt, self.ctx_cur.ctx_flatten())


    def __pop(self, eoc):
        pctx, ctx = None, self.ctx_cur
        while ctx.parent and ctx.eoc != eoc:
            pctx, ctx = ctx, ctx.parent
        if not ctx.parent: return False
        if pctx:
            pctx.parent = ctx.parent
        else:
            self.ctx_cur = ctx.parent
        return True


    def __l10n(self, txt):
        vers = txt.split('\f')
        best = (vers[0], 0.5)
        glang = self['lang'] or ('','','')
        for ver in vers[1:]:
            try:
                lang, ver = ver.split(' ',1)
            except Exception as e:
                raise ValueError(f"Invalid \\f usage '{ver}'\nExample: '\\fen_us color\\fen_gb colour'")
            slang = (lang.lower()+'___').split('_')[:3]
            
            m = 0 if slang[0] != glang[0] else\
                1 if slang[1] != glang[1] else\
                2 if slang[2] != glang[2] else\
                3
            m += reduce(lambda a,x: a + (0.0 if x else 0.1), slang, 0)
            if m > best[1]: best = (ver, m)
        return best[0]


    def rich_append(self, el, ctx):
        try:
            el = el.ctx_parent(self)
        except AttributeError:
            pass
        self.rich_stream.append((el,ctx))
        

    def rich_newline(self, ctx):
        self.rich_stream.append(('\v',ctx))
    

    def rich_tab(self, ctx):
        self.rich_stream.append(('\t',ctx))

    

def _unindent(txt):
    if '\n' not in txt: return txt
    newlines = txt.splitlines()
    indent = ''
    for line in newlines[1:]: # The first line is skipped because it starts right after the opening '''
        stripped = line.lstrip()
        if stripped: # First non-blank line sets the indentation level
            indent = line[:len(line)-len(stripped)]
            break
    n_indent = len(indent)
    if indent: newlines = [l[n_indent:] if l.startswith(indent) else l for l in newlines]
    while newlines and not newlines[-1].strip(): newlines = newlines[:-1]
    return '\n'.join(newlines).lstrip()



import pytest
from print_ext.borders import Borders,BorderDfn
from print_ext.line import SMark as SM
from print_ext.flex import Flex
from .printer_test import _printer


def test_borders_hi():
    b = Borders('hi', style='y')
    rows = [r.styled() for r in b.flatten()]
    assert(rows[0] == ('┌──┐',[SM('y',0,4), SM('dem',0,4)]))
    assert(rows[1] == ('│hi│',[SM('y',0,4), SM('dem',0,1),SM('dem',3,4)]))
    assert(rows[2] == ('└──┘',[SM('y',0,4), SM('dem',0,4)]))



def test_borders_blank():
    b = Borders(border=(1,'-'), ascii=True)
    assert([r.styled()[0] for r in b.flatten()] == ['++','++'])



def test_borderdfn_dfns():
    for v in ['aaa', 999, 'a'*7, 'a'*9, 'a'*11, 'a'*12]:
        with pytest.raises(ValueError):
            BorderDfn(t=v)
    for v in ['', 'a','aa','aaa','aaaaa']:
        with pytest.raises(ValueError):
            BorderDfn(m=v)
    for v in ['bad','also bad', 0, None]:
        print(f"BAD {v!r}")
        with pytest.raises(ValueError):
            BorderDfn(v)
    for v in [' ','-','=','{','(','[',False]:
        BorderDfn(v)



def test_borderdfn_fld_select():
    assert(BorderDfn('c::')['c'] == ':'*8)
    assert(BorderDfn(' .tc')['t'] == ' '*10)
    assert(BorderDfn(' .tc')['b'] == '\n'*10)
    with pytest.raises(ValueError):
        assert(BorderDfn('pad.t.c')['b'] == '\n'*10)



def test_border_pretty():
    o,p = _printer(color=False, ascii=True)
    f = Flex('hi\vthere\vbob')
    assert((f.width, f.height) == (5,3))
    p(Borders('hi\vthere\vbob', border=BorderDfn('-',m='01'), style='1'))
    print(o.getvalue())
    assert(repr(o.getvalue()) == "'|hi   |\\n|there|\\n|bob  |\\n'")
    
    

def test_border_size():
    b = BorderDfn(m='01')
    q = Borders('the\vquick', border=b)
    assert((b.width, b.height) == (2,0))
    assert((q.width, q.height) == (7,2))



def test_borderdfn_top_line():
    b = BorderDfn(t='abcde12345',m='1111')
    assert(b.top_line(3, True) == ' 5 ')
    assert(b.top_line(4, False) == ' ad ')
    assert(b.top_line(5, False) == ' abd ')
    assert(b.top_line(6, False) == ' abcd ')
    assert(b.top_line(7, False) == ' acbcd ')
    assert(b.top_line(8, False) == ' acbccd ')
    b = BorderDfn(t='abcde12345',b='qwert09876', c='XYZTxyzt', m='0111')
    assert(b.top_line(99, True) == None)
    assert(b.bottom_line(5, False) == 'ZqwrT')
    b = BorderDfn(t='abcde12345',c='XYZTxyzt', m='1')
    assert(b.top_line(5, False) == 'XabdY')
    assert(b.top_line(2, True) == 'xy')
    b = BorderDfn(t='abcde12345',c='XYZTxyzt', m='1101')
    assert(b.top_line(5, True) == '1234y')
    b = BorderDfn(t='abcde12345',c='XYZTxyzt', m='1110')
    assert(b.top_line(5, True) == 'x1234')
    b = BorderDfn(t='abcde12345',c='XYZTxyzt', m='1000')
    assert(b.top_line(5, True) == '13234')
    assert(b.bottom_line(5, True) == None)
    b = BorderDfn(m=1)
    assert(b.top_line(7, True) == '       ')
    


#def test_borderdfn_cascade():
    b = BorderDfn('=', 1)
    print(b)
    assert(b.top_line(5, True) == '#===#')



def test_borderdfn_override():
    b1 = BorderDfn('-', m='111\n',r=' ',c='┌─└─'*2)
    b2 = BorderDfn('-', m='11\n1',r='│|', l=' ',)
    assert(b1.update(b2).vals() == ('┌┐└┘++++', '          ', '│││││|||||', '─────-----', '─────-----', '1111'))
    assert(BorderDfn().update(BorderDfn(t='a')).vals() == ('\n'*8,'\n'*10,'\n'*10,'a'*10,'\n'*10,'\n'*4))



import pytest
from print_ext.rich import Rich
from print_ext.line import Line, SMark as SM
from print_ext.span import Span
from .testutil import ostr_ctx, debug_dump

class RichTest(Rich):
    context_defaults = dict(rt='a')

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self.calls = []
        self(*args)
    
    def rich_append(self, el, ctx):
        if isinstance(el, str): el = Line(parent=self, **ctx).insert(0,Span(el))
        self.calls.append(el)

    def rich_tab(self, ctx):
        self.calls.append('\t')

    def rich_newline(self, ctx):
        self.calls.append('\n')

    def __str__(self):
        return ','.join(map(str, self.calls))



class SpecialChars():
    def __str__(self):
        return '\b \t \n \v \f \r'



def test_rich_pop_two():
    t = RichTest(style='6-7')
    t('\b1\b2 the', 'bad', style='8-9')('bob')
    print('\n'.join(debug_dump(t)))
    assert(t.calls[0].styled() == ('the', [SM('6',0,3), SM('7',0,3), SM('8',0,3), SM('9',0,3), SM('1',0,3), SM('2',0,3)]))
    assert(t.calls[1].styled() == ('bad', [SM('6',0,3), SM('7',0,3), SM('8',0,3), SM('9',0,3)]))
    assert(t.calls[2].styled() == ('bob', [SM('6',0,3), SM('7',0,3)]))
 


def test_rich_pop_mid():
    t = RichTest(style='7')
    t('\b2\b1$ cow', 'fox', '\b4\b$ cat', style='9-8')
    assert(t.calls[0].styled() == ('cow', [SM('7',0,3), SM('9',0,3), SM('8',0,3), SM('2',0,3), SM('1',0,3)]))
    assert(t.calls[1].styled() == ('fox', [SM('7',0,3), SM('9',0,3), SM('8',0,3), SM('1',0,3)]))
    assert(t.calls[2].styled() == ('cat', [SM('7',0,3), SM('9',0,3), SM('8',0,3), SM('4',0,3)]))



def test_rich():
    t = RichTest()#rt='b')
    #assert(t['rt'] == 'a')
    assert(str(t('bob')) == 'bob')
    assert(str(RichTest()('\t')('', '\v')) == '\t,\n')



@pytest.mark.skip(reason="Not implemented")
def test_l10n():
    t = 'xxx\fa_b_c abc\fa a\fa_b ab\fa_q aq'
    assert(str(RichTest(t)) == 'xxx')
    assert(str(RichTest(t,lang='a')) == 'a')
    assert(str(RichTest(t,lang='a_b')) == 'ab')
    assert(str(RichTest(t,lang='a_b_c')) == 'abc')
    assert(str(RichTest(t,lang='A_Q')) == 'aq')
    assert(str(RichTest(t,lang='a_q_d')) == 'aq')
    assert(str(RichTest(t,lang='a_b_q')) == 'ab')
    assert(str(RichTest(t,lang='a_f')) == 'a')
    assert(str(RichTest(t,lang='c')) == 'xxx')
    assert(str(RichTest(t,lang='c_b_c')) == 'xxx')



def test_append_tel():
    p = RichTest(z='cob')
    a = RichTest('hi', parent=p)
    assert(a['z'] == 'cob')
    b = RichTest(a, z='bob')
    assert(b.calls[0]['z'] == 'cob') # a  shouldn't get reparented to b



def test_rich_call_kwargs():
    r = RichTest('0','1', a=3, m=4)
    r('2', a=9, q=10)
    assert(r.calls[0]['a'] == 3)
    assert(r.calls[0]['q'] == None)
    assert(r.calls[2]['a'] == 9)
    assert(r.calls[2]['q'] == 10)
    r('3')
    assert(r.calls[3]['a'] == 3)
    assert(r.calls[0]['q'] == None)
    assert(r.calls[0]['m'] == 4)



def test_append_non_string():
    special_chars = SpecialChars()
    assert(str(RichTest(1,2,[3,4])) == f'1,2,{[3,4]}')
    assert(str(RichTest(special_chars).calls[0]) == '\\x08      \\x0a \\x0b \\x0c \\x0d')



def test_cleandoc():
    assert(str(RichTest('''   The quick
        brown
        fox
    
    ''')) == 'The quick,\n,brown,\n,fox')
    assert(str(RichTest('a\n  b\n  c\v\n\n\n')) == 'a,\n,b,\n,c')
    assert(str(RichTest('\n  a\n  b\n  c\v\n\n\n')) == 'a,\n,b,\n,c')
    


def test_tabs():
    assert(str(RichTest('a\tb\v\tc\t\v')) == 'a,\t,b,\n,\t,c,\t,\n')
    assert(str(RichTest('a\n\t\tb\tc\n\t\td\n\t\t\te\n\t\t')) == 'a,\n,b,\t,c,\n,d,\n,\t,e')
    assert(str(RichTest('\t\n\ta \n\tb\t\n\tc\t\t\n\t\n\t\t\t')) == 'a ,\n,b,\t,\n,c,\t,\t')



def test_rich_attr():
    a = RichTest('\bR\bB a')
    assert(str(a) == 'a')
    assert(a.calls[0]['style'] == ('R','B'))
    a = RichTest('\bR \br \bpi=3.4 a')
    assert(str(a) == 'a')
    assert(a.calls[0]['style'] == ('R','r'))
    assert(a.calls[0]['PI'] == 3.4)



def test_rich_pop():
    a = RichTest('\ba=1 \ba=2 bob \b the \b cob', a=10)
    assert(a.calls[0]['a'] == 2)
    assert(a.calls[1]['a'] == 1)
    assert(a.calls[2]['a'] == 10)



def test_autopop_call():
    a = RichTest('\ba=1$','b')
    a('2')
    assert(a.calls[0]['a'] == 1)
    assert(a.calls[1]['a'] == None)



def test_autopop_str():
    a = RichTest('\ba=1 a','b')
    assert(a.calls[0]['a'] == 1)
    assert(a.calls[1]['a'] == None)



def test_multi_style():
    r = RichTest('\bx$ hi', '\vbye', '\v\by bob', 'coj', style='z')
    r('jim')
    a = r.calls
    print(a)
    assert(a[0]['style'] == ('z','x'))
    assert(a[2]['style'] == ('z','x'))
    assert(a[4]['style'] == ('z','x','y'))
    assert(a[5]['style'] == ('z','x'))
    assert(a[6]['style'] == ('z',))





import pytest 
from functools import partial
from print_ext.size import Size


def calc(size, *args, elide=None, **kwargs):
    def _size(arg):
        # (rate, min)
        # (rate, min, max)
        # (rate, min, max, nom)
        # (nom, min, max, rate)
        if not isinstance(arg, tuple): return Size(rate=arg)
        elif len(arg) == 2: return Size(rate=arg[0], min=arg[1])
        elif len(arg) == 3: return Size(rate=arg[0], min=arg[1], max=arg[2])
        elif len(arg) == 4: return Size(rate=arg[0], min=arg[1], max=arg[2], nom=arg[3])
        
        
    els = [_size(a) for a in args]
    rstr = ''
    tot, ftot = 0, []
    rows = Size.resize(els, size, elide=elide, **kwargs)
    print(rows)
    for ri, row in enumerate(rows):
        rmax = sum([c.max for c in row])
        rmin = 0 if elide else sum([c.min for c in row])
        rsize = sum([c.size for c in row])
        print(f"{ri}: min:{rmin}/ask:{size}/max:{rmax}/got:{rsize} -- {row}")
        if size: assert(rsize == max(rmin, min(rmax, size)))
        for c in row:
            if not elide: assert(c.size >= c.min - Size.err)
            assert(c.size <= c.max + Size.err)
        rstr += ''.join(f' {c.size}' for c in row) + ' --'
    return rstr[:-3].strip()


def test_checks():
    Size()
    Size(min=0, max=0)
    Size(rate=0.1)
    for args in [{'rate':-1}, {'min':0.5, 'max':0.499999}]:
        with pytest.raises(ValueError):
            Size(**args)
   

def test_size_one():
    C = partial(calc, 100, wrap=False)
    assert(C((1,10,500,86.66666666666667)) == '100')

    for args in [(1,0,200,0), (1,0,200,120), (1,0,200,-100), 4, 0.1, (1, 99.9), (1,1,100.1)]:
        print(args)
        assert(C(args) == '100')
    assert(C((1,0,0,100)) == '0')
    assert(C((1,120)) == '120')
    assert(C((1,5,99)) == '99')
    

@pytest.mark.xfail
def test_size_rate0():
    ''' Zero rate fixes the size to nominal '''
    raise Exception()



def test_two():
    C = partial(calc, 100, wrap=False)
    assert(C((2000,60), (1, 70)) == '60 70')
    assert(C((2000,1,10), (1, 1, 60)) == '10 60')
    assert(C(1, 1) == '50 50')
    assert(C(2, 1) == '67 33')
    assert(C(200000, 1) == '99 1')
    assert(C((2000,1), (1, 25)) == '75 25')
    assert(C((2000,1,10), (1, 25)) == '10 90')
    assert(C((1,60,60), 1) == '60 40')



def test_three():
    C = partial(calc, 100, wrap=False)
    assert(C((1,40,40), 1, 1) == '40 30 30')
    assert(C((1,40,40), 2, 1) == '40 45 15')
    assert(C((1,40,40,40), 2, 1) == '40 40 20')


def test_zero():
    C = partial(calc, 0, wrap=False)
    assert(C((1,40,40), (1,0,10000, 10), (1,0,20,10000)) == '40 10 20')


def test_size_wrap():
    C = partial(calc, 100, wrap=True)
    assert(C((1, 60, 80, 60), (1, 50, 500, 1), (1, 10, 500, 1)) == '80 -- 50 50')


def test_size_round():
    ''' rounding to integer boundries produces a +/- 1 problem '''
    C = partial(calc, 20, wrap=True)
    assert(C((1.0,0,5,9), (1.0,0,10000,3), (1.0,0,10000,1)) == '5 9 6')


def test_size_fractional():
    ''' Fractionally small cells may round to zero.  If you set a minimum of zero then you have to deal with it. '''
    C = partial(calc, 3, wrap=False)
    assert(C((4,0), (1,0), (4,0), (1,0), (4,0)) == '1 0 1 0 1')


def test_overflow():
    ''' A Single cell overflows the width so it must be truncated with a warning.
    This is the only situation where the cell will fall out of bounds (less than min)
    '''
    def elide(x):
        return Size(min=4, max=8)
    assert(calc(100, (1,120), 1, (1, 99), (1,101), wrap=True, elide=(1,4,8)) == '100 -- 1 99 -- 100')
    assert(calc(100, (1,120), 1, (1, 99), (1,101), wrap=False, elide=elide) == '96 4')
    assert(calc(100, (1,120), 100, (1, 99), (1,101), wrap=True, elide=None) == '120 -- 1 99 -- 101')
    assert(calc(100, (1,120), 1, (1, 99), (1,101), wrap=False, elide=None) == '120 1 99 101')


def test_elide_zero():
    ''' If a cell wants to be zero don't force an elision '''
    def die(x):
        assert(False)
    assert(calc(100, 1, (1, 0, 0), wrap=True, elide=die) == '100 0')
    

def test_auto_params():
    ''' dont set a minimum < nom for nom < 12.
        Set min as size/4 for nom >=12
    '''
    assert(Size(nom=8).min == 8)
    assert(Size(nom=12).min == 3)
    s = Size(nom=3, rate=None)
    assert( (s.min, s.max) == (3, 3))
    assert(Size(nom=30, min=0).min == 0)
    s = Size(nom=0, min=0, max=0, rate=1)
    assert( (s.min, s.max) == (0,0))
    assert(Size(nom=0).min == 1)
    assert(Size(nom=5, max=2).min == 2)

if __name__=='__main__':
    test_overflow()



from .context_cvar import ctx_vars, CallableVar, ObjectAttr, CVar, IntCVar, FloatCVar, BoolCVar, EnumCVar


class MetaContext(type):
    def __new__(self, name, bases, attrs, **kwargs):
        cls = super().__new__(self, name, bases, attrs)
        try:
            cls.ctx_class = dict(cls.__mro__[1].ctx_class)
            cls.ctx_class_trace = list(cls.__mro__[1].ctx_class_trace)
        except:
            cls.ctx_class = {}
            cls.ctx_class_trace = []
        cvars = ctx_vars()
        cls.ctx_class_trace.insert(0, {cvars[k]:v if isinstance(v, CallableVar) else cvars[k].canon(v) for k,v in kwargs.items()})
        for cv,v in cls.ctx_class_trace[0].items():
            if hasattr(cv, 'merge') and cv.names[0] in cls.ctx_class:
                v = cv.merge(v, cls.ctx_class[cv.names[0]])
            cls.ctx_class[cv.names[0]] = v
        return cls



class Context(metaclass=MetaContext):
    ''' 
    The lookup for a context variable
     1. defined on self
     2. defined in a parent
     3. defined in ctx_defaults
     4. defined in a parent's ctx_defaults
     5. None

    A context object might want to live in two places at once::

        obj = Bob()
        Cob(obj)
        Bill(obj)
        obj['x'] == ???

    Or you might create loops::

        a = Apple()
        b = Ball()
        a(b)
        b(a)
    '''


    #@classmethod
    #def defaults(self, **kwargs):
    #    print(f"DEFAULTS {self}  {kwargs}")
    #    def _f(cls):
    #        cvars = ctx_vars()
    #        cls._clsctx = {cvars[k].names[0]: v if isinstance(v, CallableVar) else cvars[k].canon(v)  for k,v in kwargs.items()}
    #        return cls
    #    return _f


    @classmethod
    def define(self, cvar):
        cvars = ctx_vars()
        for name in cvar.names:
            if name not in cvars:
                cvars[name] = cvar
            elif cvar != cvars[name]:
                raise ValueError(f"Can't register the duplicate name '{name}'\nCurrent: {cvars[name]}\nTrying to register: {cvar}")
        return cvar


    def __init__(self, parent=None, **kwargs):
        self.ctx_local = {}
        self.children = set()
        self.parent = parent
        if parent != None: parent.children.add(self)
        for k,v in kwargs.items():
            try:
                self.ctx(**{k:v})
            except KeyError:
                pass
        self.changed_size()


    def ctx(self, *keys, **set_vals):
        ''' get/set/delete context variables on this object

        If you want to override __getitem__/__setitem__ to do other things then you can use this method to access the context.

        Parameters:
            *keys (str,)
                Key names to lookup.  They may be aliases.
            **set_vals
                Set these values on this object.  If the key is also defined in ``keys`` then
                the previous value will also be returned.
                
                Setting a key to None clears the value from self.

        Returns:
            A tuple of values, one for each key in ``*keys``.

        Raises:
            KeyError
                A key name has not been defined with ``Context.define()``
        '''
        cvars = ctx_vars()
        vals = (self._ctx_lookup(cvar) for cvar in (cvars[k] for k in keys))
        vals = tuple(v(self) if isinstance(v, CallableVar) else v for v in vals)
        for k, v in set_vals.items():
            cv = cvars[k]
            k = cv.names[0]
            if v != None:
                v = v if isinstance(v, CallableVar) else cv.canon(v)
                self.ctx_local[k] = v
            elif k in self.ctx_local:
                del self.ctx_local[k]
        if set_vals: self.changed_ctx()      
        return None if len(vals) == 0 else vals[0] if len(vals) == 1 else vals


    def changed_ctx(self):
        self.changed_size()
        for child in self.children:
            child.changed_ctx()


    def changed_size(self):
        pass


    def each_child(self):
        yield from []


    def ctx_contains(self, other):
        ctx = self
        while ctx != None:
            if id(ctx) == id(other): return True
            ctx = ctx.parent
        return False


    def ctx_parent(self, parent):
        el = self
        if parent!=None and parent.ctx_contains(self) or self.parent != None and id(self.parent) != id(parent):
            el = self.clone(**self.ctx_flatten())
        el.parent = parent
        if parent!=None: parent.children.add(el)
        return el


    def ctx_trace(self):
        cvars = ctx_vars()
        return [{cvars[k]:v for k,v in self.ctx_local.items()}, *self.__class__.ctx_class_trace]


    def clone(self, *args, **kwargs):
        ''' Perform a deep clone of this object.
        '''
        return self.__class__(*args, **kwargs)


    def _ctx_lookup(self, cvar):
        try:
            v = self._ctx_lookup_merge(cvar.merge, cvar.names[0])
        except AttributeError:
            v = self._ctx_lookup_nomerge(cvar.names[0])[0]
        return v #v(self) if isinstance(v, CallableVar) else v


    def _ctx_lookup_nomerge(self, key):
        # No merge needed so just pick the closest value
        try:
            return self.ctx_local[key], True # self object has defined a value
        except KeyError: pass
        # Our object does not define cvar,  Check parent
        try:
            val_src = self.parent._ctx_lookup_nomerge(key)
            if val_src[1]: return val_src # Our parent context is defining a value
            # Check if we have a better default to use
        except AttributeError:
            # Parent may not be defined, or may not be a Context object with ctx_lookup
            val_src = None, None # This is our last ditch effort
        # If our class defines cvar then that is better than the parent's class cvar or None
        return (self.ctx_class.get(key, None), None) or val_src


    def _ctx_lookup_merge(self, merge, key):
        vals = []
        obj = self
        while obj != None:
            vals.insert(0, (obj.ctx_class.get(key, None), obj.ctx_local.get(key, None)))
            try: obj = obj.parent
            except AttributeError: obj = None
        val = None
        for v in vals: val = v[0] if val == None else val if v[0] == None else merge(v[0], val)
        for v in vals: val = v[1] if val == None else val if v[1] == None else merge(v[1], val)
        return val


    def ctx_flatten(self):
        ''' Return a complete dictionary of context variables
        '''
        vals = ((cvar.names[0], self._ctx_lookup(cvar)) for cvar in set(ctx_vars().values()))
        return {k:v for k,v in vals if v != None}
        

    def __getitem__(self, key):
        return self.ctx(key)


    def __setitem__(self, key, val):
        self.ctx(**{key:val})


    def __delitem__(self, key):
        self.ctx(**{key:None})



from functools import reduce



class BorderDfn():
    ''' Define the border by single characters

    \n don't care
    corners      [c] : TL TR BL BR
    mask         [m] : T B L R
    top/bottom [t/b] : X | LXR | LMXR | LMXR1
    left/right [l/r] : X | TXB | LMXR | TMXB1

    '''
    special = {
        '\\':'╲', '/':'╱', 'X':'╳',
        '|2':'╎', '-2':'╌', '#|2':'╏', '#-2':'╍',
        '|3':'┆', '-3':'┄', '#|3':'┇', '#-3':'┅',
        '|4':'┊', '-4':'┈', '#|4':'┋', '#-4':'┉',
    }

    codes = {
        '   #':'╸', '   -':'╴', '  # ':'╻', '  ##':'┓', '  #-':'┒', '  - ':'╷', '  --':'┐', '  -=':'╕', '  =-':'╖',
        '  ==':'╗', '  ~~':'╮', ' #  ':'╺', ' # #':'━', ' # -':'╾', ' ## ':'┏', ' ###':'┳', ' ##-':'┲', ' #- ':'┍',
        ' #-#':'┯', ' #--':'┮', ' -  ':'╶', ' - -':'─', ' -# ':'┎', ' -##':'┱', ' -- ':'┌', ' --#':'┭', ' ---':'┬',
        ' -= ':'╓', ' -=-':'╥', ' = =':'═', ' =- ':'╒', ' =-=':'╤', ' == ':'╔', ' ===':'╦', ' ~~ ':'╭', '#   ':'╹',
        '#  #':'┛', '#  -':'┚', '# # ':'┃', '# ##':'┫', '# #-':'┨', '# - ':'╿', '# -#':'┩', '# --':'┦', '##  ':'┗',
        '## #':'┻', '## -':'┺', '### ':'┣', '####':'╋', '###-':'╊', '##- ':'┡', '##-#':'╇', '##--':'╄', '#-  ':'┖',
        '#- #':'┹', '#- -':'┸', '#-##':'╉', '#-#-':'╂', '#-- ':'┞', '#--#':'╃', '#---':'╀', '-   ':'╵', '-  #':'┙',
        '-  -':'┘', '-  =':'╛', '- # ':'╽', '- ##':'┪', '- #-':'┧', '- - ':'│', '- -#':'┥', '- --':'┤', '- -=':'╡',
        '-#  ':'┕', '-# #':'┷', '-# -':'┶', '-###':'╈', '-##-':'╆', '-#- ':'┝', '-#-#':'┿', '-#--':'┾', '--  ':'└',
        '-- #':'┵', '-- -':'┴', '--# ':'┟', '--##':'╅', '--#-':'╁', '--- ':'├', '---#':'┽', '----':'┼', '-=  ':'╘',
        '-= =':'╧', '-=- ':'╞', '-=-=':'╪', '=  -':'╜', '=  =':'╝', '= = ':'║', '= =-':'╢', '= ==':'╣', '=-  ':'╙',
        '=- -':'╨', '=-= ':'╟', '=-=-':'╫', '==  ':'╚', '== =':'╩', '=== ':'╠', '====':'╬', '~  ~':'╯', '~~  ':'╰',
    }

    codes_inv = dict(zip(codes.values(), codes.keys()))    
    fields = 'clrtbm'
    fld_size = dict(c=8,l=10,r=10,t=10,b=10,m=4)
    dfns = {}


    @staticmethod
    def ext(trbl):
        return ''.join(BorderDfn.codes_inv.get(c,'    ')[i] for i, c in enumerate(trbl))


    @staticmethod
    def join(code=None, t=' ', r=' ', b=' ', l=' '):
        code = code or t+r+b+l
        return BorderDfn.codes.get(code, BorderDfn.special.get(code, ' '))


    @staticmethod
    def define(name, *args, **kwargs):
        dfn = BorderDfn(*args, **kwargs)
        if name in BorderDfn.dfns and BorderDfn.dfns[name] != dfn:
            raise ValueError(f"Redefining BorderDfn {name!r}:\nExisting: {BorderDfn.dfns[name]}\nTrying to set: {dfn}")
        BorderDfn.dfns[name] = dfn


    @staticmethod
    def _to_bdr(v):
        if isinstance(v, BorderDfn): return v.clone()
        v = str(v).lower().split(':',1)#'.',1)
        if len(v) == 1:
            v = v[0].split('.',1)
            if v[0] not in BorderDfn.dfns: raise ValueError(f"{v[0]!r} is not a defined BorderDfn")
            bdr = BorderDfn.dfns[v[0]]
            try:
                return bdr if len(v) == 1 else BorderDfn(**{k:bdr[k] for k in v[1]})
            except KeyError:
                raise ValueError(f"Illegal selector {v[1]!r} in {'.'.join(v)!r}") 
        else:
            return BorderDfn(**{v[0]:v[1]})


    def __init__(self, *vals, **kwargs):
        self._sides = None
        if vals:
            base = reduce(lambda a,b: a.update(b), [BorderDfn._to_bdr(b) for b in vals])
        else:
            base = {k:'\n' for k in BorderDfn.fields}
        for k in BorderDfn.fields:
            self[k] = str(kwargs.get(k, base[k]))


    def __getitem__(self, k):
        try:
            return getattr(self, k)
        except AttributeError:
            return '\n'*BorderDfn.fld_size[k]


    def __setitem__(self, k, v):
        if len(v) == 1: v = v*BorderDfn.fld_size.get(k,2)
        if k == 'tl': k,v = 'c', v[0]+'\n\n\n'+v[1]+'\n\n\n'
        elif k == 'tr': k,v = 'c', '\n'+v[0]+'\n\n\n'+v[1]+'\n\n'
        elif k == 'bl': k,v = 'c', '\n\n'+v[0]+'\n'+v[1]+'\n'
        elif k == 'br': k,v = 'c', '\n\n\n'+v[0]+'\n\n\n'+v[1]
        if len(v) == 2: v = v[0]*(BorderDfn.fld_size[k]//2) + v[1]*(BorderDfn.fld_size[k]//2)
        if len(v) * 2 == BorderDfn.fld_size[k]: v = v+v
        if k == 'm':
            if not all(c in '01\n' for c in v): raise ValueError(f"Invalid mask value: {k}={v!r}")
        elif k in 'trbl':
            if len(v) == 6: v = v[:2] + v[1:3] + v[1] + v[3:5] + v[4:6] + v[4]
            elif len(v) == 8: v = v[:4] + v[1] + v[4:] + v[5]
        if len(v) != BorderDfn.fld_size[k]: raise ValueError(f"Invalid BorderDfn: {k}={v!r}")
        setattr(self, k, ''.join(o if n=='\n' else n for o,n in zip(self[k], v)))


    def clone(self):
        return BorderDfn(**{k:self[k] for k in BorderDfn.fields})


    def vals(self):
        return tuple(self[k] for k in BorderDfn.fields)


    def update(self, newer):
        out = self.clone()
        for k in BorderDfn.fields:
            out[k] = newer[k]
        return out


    def __eq__(self, other):
        return self.vals() == (other and other.vals())


    def __hash__(self):
        return hash(self.vals())


    def __repr__(self):
        _m = lambda v: f'{v[0]!r}*{len(v)}' if v[0]*len(v) == v else repr(v)
        args = [f'{k}={_m(self[k])}' for k in BorderDfn.fields]
        return f"BorderDfn({','.join(args)})"


    @property
    def sides(self):
        if not self._sides:
            self._sides = [c=='1' for c in self.m]
        return self._sides
    

    @property
    def width(self):
        _, _, l, r = self.sides
        return int(l) + int(r)


    @property
    def height(self):
        t, b, _, _ = self.sides
        return int(t) + int(b)


    def _line(self, w, ascii, top):
        if not self.sides[0 if top else 1]: return None
        c = self.c[4:] if ascii else self.c
        cl, w = (c[0 if top else 2], w-1) if self.sides[2] else ('', w)
        cr, w = (c[1 if top else 3], w-1) if self.sides[3] else ('', w)
        chars = self.t if top else self.b
        return (cl + self.side_chars(w, chars[5:] if ascii else chars) + cr).replace('\n',' ')


    def top_line(self, w, ascii):
        return self._line(w, ascii, top=True)


    def bottom_line(self, w, ascii):
        return self._line(w, ascii, top=False)


    def side_chars(self, w, chars):
        if not w: return ''
        mid = w//2 + 1
        txt, w = (chars[4], w-1) if w == 1 else (chars[0], w-1)
        while w:
            if w == 1: txt += chars[3]
            elif w == mid: txt += chars[1]
            else: txt += chars[2]
            w -= 1
        return txt



BorderDfn.define('false', m='0')
BorderDfn.define('1', m='1')
BorderDfn.define(' ', c=' ', t=' ', b=' ', l=' ', r=' ')
BorderDfn.define('-', c='┌┐└┘++++', t='─-', b='─-', l='│|', r='│|')
BorderDfn.define('=', c='╔╗╚╝####', t='═=', b='═=', l='║#', r='║#')
BorderDfn.define('[', t='┌──┐⎴-----', b='└──┘⎵-----', l='⎡⎢⎢⎣[[[[[[', r='⎤⎥⎥⎦]]]]]]')
BorderDfn.define('(', t='╭──╮⏜/--\\-', b='╰──╯⏝\\--/-', l='⎛⎜⎜⎝(/||\\(', r='⎞⎟⎟⎠)\\||/)')
BorderDfn.define('{', t='╭┴─╮⏞/`-\\*', b='╰┬─╯⏟\\,-/*', l='⎧⎨⎪⎩{/{|\\{', r='⎫⎬⎪⎭}\\}|/}')
BorderDfn.define('box', '1', '-')



import copy, sys
from math import ceil
from functools import reduce
from collections import namedtuple
from .span import Span
from .context import Context, CVar, BoolCVar, IntCVar
from .rich import Rich


class Just():
    __slots__ = ('hv','h','v')

    def __init__(self, hv=None, defaults=''):
        if not hv:
            hv = ('','')
        elif isinstance(hv, str):
            if len(hv) == 1:
                hv = ('',hv) if hv in '_-~^' else (hv,'')
            else:
                hv = tuple(hv)
        elif isinstance(hv, Just):
            hv = (hv.h, hv.v)
        if defaults:
            defaults = Just(defaults)
            hv = (hv[0] or defaults.h, hv[1] or defaults.v)
        self.h, self.v = hv
        if self.h not in '<|:>': raise ValueError(f"Invalid horizontal justification character '{self.h}'")
        if self.v not in '_-~^': raise ValueError(f"Invalid vertical justification character '{self.v}'")
        


    def pad_v(self, h):
        if self.v == '_': return h
        elif self.v == '-': return h//2
        elif self.v == '~': return ceil(h/2)
        return 0


    def pad_h(self, w):
        if self.h == '>': return w
        elif self.h == '|': return w//2
        elif self.h == ':': return ceil(w/2)
        return 0
        

    def __str__(self):
        return self.h+self.v


    def __repr__(self):
        return f'Just({self.h+self.v!r})'



class JustCVar(CVar):
    def canon(self, sval):
        return Just(sval)

    def merge(self, val, pval):
        return Just(val, pval)



class StyleCVar(CVar):
    def canon(self, sval):
        if isinstance(sval, str):
            if ' ' in sval: raise ValueError(f"No spaces allowed in style string '{sval}'")
            if not sval: return None
            sval = tuple(sval.split('-'))
        return sval[-1 - list(reversed(sval)).index(''):] if '' in sval else sval


    def merge(self, val, pval):
        if '' in val: return val[-1 - list(reversed(val)).index(''):]
        # find overlap
        for same in range(min(len(pval),len(val)), 0, -1):
            if pval[-same:] == val[:same]:
                return pval + val[same:]
        return pval + val
        

style_cvar = Context.define(StyleCVar('style',''))
Context.define(BoolCVar('ascii', 'ascii_only'))
Context.define(BoolCVar('text_wrap', 'wrap'))
Context.define(JustCVar('justify', 'just'))
Context.define(IntCVar('wrap_mark_from', 'wmf'))



class SMark():
    __props__ = ('style','s','e')

    def __init__(self,*args):
        self.style, self.s, self.e = args

    def __repr__(self):
        return f'{self.style}/{self.s}-{self.e}'

    def __eq__(self, other):
        return self.style == other.style and self.s == other.s and self.e == other.e



class Line(Rich, wrap=True):
    ''' A Line is a list of Lines and Spans.
    
    It is a Context so it can style the text.
    Only Lines and Spans can be appended (or inserted).
    Anything else will be converted with Span() first and then added.
    Any Lines appended or inserted into this string will become Context children.
    '''
    
    def __init__(self, *args, **kwargs):
        self.__spans = []
        super().__init__(*args, **kwargs)


    @property
    def spans(self):
        while self.rich_stream:
            el, ctx = self.rich_stream.pop(0)
            if el == '\v': el = '\\n'
            self.insert(-1, el, **ctx)
        return self.__spans


    def trim(self, width):
        ''' Trim this string to fit into width

        If width >= 0 then keep the left
        If width < 0 then keep the right
        '''
        w = self.width+width if width < 0 else width
        for i, span in enumerate(self.spans):
            if span.width > w: break
            w -= span.width
        else:
            return self
        span = span.trim(-span.width+w) if width < 0 else span.trim(w)
        self.__spans[slice(0,i+1) if width < 0 else slice(i,len(self.__spans))] = [span] if span else []
        return self


    def calc_width(self):
        return reduce(lambda a,x: x.width + a, self.spans, 0)
    

    def calc_height(self):
        return int(self)


    def each_child(self):
        yield from self.spans


    def justify(self, width, justify_h=''):
        ''' Add padding so that we are correctly justified for the given `width`
        '''
        if width == 0: return self
        pad = width - self.width
        if pad < 0: raise ValueError(f"This string width ({self.width}) is greater than the given justification width ({width})")        
        padl = Just(justify_h, Just(self['justify'], '<')).pad_h(pad)
        return self.insert(0, ' '*padl).insert(-1, ' '*(pad-padl))



    def insert(self, at, el, **kwargs):
        if isinstance(el, Line):
            pass
        elif hasattr(el, 'flatten'):
            flat = list(el.flatten())
            if not flat: return self
            if at == 0: flat.reverse()
            self.insert(at, flat[0], **kwargs)
            for el in flat[1:]:
                self.insert(at, '\\n', **kwargs)
                self.insert(at, el, **kwargs)
            return self
        else:
            el = Span(el)
        if not el: return self
        if kwargs: el = Line(**kwargs).insert(0, el)
        el = el.ctx_parent(self)
        if self.__spans and isinstance(el, Span) and isinstance(self.__spans[at], Span):
            self.__spans[at] = Span(self.__spans[-1], el) if at else Span(el, self.__spans[0])
        elif at:
            self.__spans.append(el)
        else:
            self.__spans.insert(0, el)
        self.changed_size()
        return self


    def clone(self, **kwargs):
        s = self.__class__(**kwargs)
        s.__spans = [x if isinstance(x,Span) else x.clone(parent=s, **x.ctx_local) for x in self.spans]
        return s


    def styled(self):
        ''' Returns a str and a list of style commands (push/pop)
        '''
        tstk = [] # temporary stack
        stack = [] # progressive stack
        
        def _stack_append(style, s, e):         
            if not style: return   
            popped = []
            while tstk and tstk[-1].e <= s: popped.insert(0,tstk.pop()) # Pop old ones
            #print(f"## {''.join(style)}:{s}-{e} {tstk} :: {popped}")
            # Match as many parents as we can
            for same in range(len(style), -1, -1):
                if style[:same] == tuple(y.style for y in tstk[-same:]): break
            style = style[same:]
            #print(f"    {same} matched parents")
            # Extend popped styles
            i = 0
            for p, y in zip(popped, style[:]):
                if p.style != y: break
                if p.e != s: break
                i += 1
                p.e = e
                tstk.append(p)
                style = style[1:]
            #print(f"    {i} extended")
            #print(f"    {style}  new")
            for y in style:
                tstk.append(SMark(y, s, e))
                stack.append(tstk[-1])


        def _styled(part, s, parent, styles):
            if isinstance(part, Line) and (y:=part.ctx_local.get('style',None)):
                styles = style_cvar.merge(y, styles) if styles else y
            if isinstance(part, Span):
                _stack_append(styles, len(s), len(s)+len(part))
                s += str(part)
            else:
                for sub in part.spans: s = _styled(sub, s, part, styles)
            return s

        return _styled(self, '', self, self['style']), stack


    def changed_size(self):
        self.__len = None
        super().changed_size()


    def flatten(self, w=0, h=0, **kwargs):
        if self.width - (w or 1e99) <= 0:
            rows = [self.clone(parent=self).justify(w)]
        else:
            rows = self._flatten_wrap(w) if w and self['text_wrap'] else self._flatten_no_wrap(w)
        dh = len(rows) - (h or 1e99)
        if dh > 0: # We are overflowing vertically by dh
            vbar = '|' if self['ascii'] else '⋮'
            keep = len(rows) - dh - 1
            n = str(dh+1)
            e = Line(f"{vbar}{n} lines{vbar}", style='dem', parent=self)
            if e.width > w: e = Line(style='dem', parent=self).insert(0, f"{vbar}{n}")
            if e.width > w: e = Line(style='dem', parent=self).insert(0, f"{vbar}")
            rows = rows[:keep-keep//2] + [e.justify(w, '|')] + rows[len(rows)-keep//2:]
        yield from justify_v(rows, h, Just(self['justify'], '^'), Line(parent=self).insert(0,' '*w))


    def _flatten_no_wrap(self, w):
        lhs = self.clone().trim(w//2)
        rhs = self.clone().trim(-((w-1)//2))
        inter = Line(('~' if self['ascii'] else '…')*(w-lhs.width-rhs.width), style='dem')
        return [Line(lhs, inter, rhs, parent=self)]
        

    def _flatten_wrap(self, w):
        cur, rows, wmf = self.clone(parent=self), [], max(self['wrap_mark_from'] or 4, 4)
        prefix = '' if w < wmf else '\\ ' if self['ascii'] else '⤷ '
        while True:
            lhs = cur.clone(parent=self).trim(w)
            cur.trim(-(cur.width - lhs.width))  
            rows.append(lhs.insert(-1, ' '*(w-lhs.width)))
            if not cur: break
            l = Line(style='dem').insert(0,prefix)
            if prefix: cur.insert(0, l)
            
        return rows


    def __str__(self):
        return ''.join(map(str, self.spans))


    def __repr__(self):
        return f"Line({', '.join([repr(x) for x in self.spans])})"
        

    def __bool__(self):
        return bool(self.spans)


    def __len__(self):
        if self.__len == None:
            self.__len = reduce(lambda a,x: len(x) + a, self.spans, 0)
        return self.__len
        


def justify_v(rows, h, j, line):
    if not h:
        yield from rows
        return
    pad = h - len(rows)
    if pad < 0: raise ValueError(f"{len(rows)} is greater than the given justification height ({h})")
    padt = j.pad_v(pad)
    for _ in range(padt): yield line.clone(parent=line.parent, **line.ctx_local)
    yield from rows
    for _ in range(pad-padt): yield line.clone(parent=line.parent, **line.ctx_local)



import types


class _CacheDecorator():
    def __init__(self, fn):
        self.fn = fn
        
    def __set_name__(self, owner, name):
        fn = self.fn.fget if isinstance(self.fn, property) else self.fn
        def _get(s, *args, **kwargs):
            try:
                return s._cache[name]
            except KeyError:
                v = fn(s, *args, **kwargs)
                s._cache[name] = v
                return v

        def _clear(s):
            if name in s._cache: del s._cache[name]

        setattr(owner, f'_clear_{name}', _clear)
        if isinstance(self.fn, property):
            setattr(owner, name, property(_get).setter(self.fn.fset).deleter(self.fn.fdel))
        elif isinstance(self.fn, types.FunctionType):
            setattr(owner, name, _get)
        else:
            raise ValueError(f"Can't cache '{name}' of type '{type(self.fn)}' on class '{owner}'")
        if not hasattr(owner, '_has_cache'):
            super_init = owner.__init__
            def _init(self, *args, **kwargs):
                self.__dict__['_cache'] = {}
                super_init(self, *args, **kwargs)  
            setattr(owner, '__init__', _init)
            owner._has_cache = True
        

    def __getattr__(self, attr):     
        def _f(*args, **kwargs):
            self.fn = getattr(self.fn, attr)(*args, **kwargs)
            return self
        return _f
        

def cache(fn):
    return _CacheDecorator(fn)



from print_ext.cache import cache


class CacheTst():
    
    @cache
    def x(self, a):
        return a


    @cache
    @property
    def y(self):
        self._y += 1
        return self._y

    @y.setter
    def y(self, v):
        self._y = v

    @property
    def z(self):
        print('getz')
    
    @z.setter
    def z(self, v):
        print(f'setz {v}')


class Desc():
    def __init__(self, name):
        self.name = name

    def __get__(self, isntance, owner=None):
        print(f"{self.name} : GET {isntance} {owner}")
        return f"Desc-{self.name}"

    def __set__(self, instance, value):
         print(f"{self.name} : SET {instance} {value}")



class Parent():
    pass

class Jim(Parent):
    
    @cache
    def __len__(self):
        return 0
    


def test_len():
    t = Jim()
    l = len(t)
    t._clear___len__()


def test_cache():
    t = CacheTst()
    t._clear_x()
    t._clear_x()
    assert(t.x(3) == 3)
    assert(t.x(4) == 3)
    t._clear_x()
    assert(t.x(4) == 4)
    assert(t.x(6) == 4)

def test_prop():
    t = CacheTst()
    print(dir(t))
    t._clear_y()
    t.y = 0
    assert(t.y == 1)
    assert(t.y == 1)
    t.y = 9
    assert(t.y == 1)
    t._clear_y()
    assert(t.y == 10)
    assert(t.y == 10)
    

if __name__ == '__main__':
    test_prop()



from math import ceil, modf
from functools import reduce
from collections import namedtuple
import logging

log = logging.getLogger('print-ext.size')
_err = 0.0001


class Size():
    __ports__ = ('nom','size','min','max','rate','user')

    err = _err
    def __init__(self, **kwargs):
        if 'easy' in kwargs:
            v = kwargs.pop('easy')
            if isinstance(v, int):
                kwargs = dict(min=-v, max=-v, rate=0) if v < 0 else dict(min=v)
            else:
                kwargs = dict(rate=v)
        args = dict(max=1e10, rate=1, nom=0, size=0, user=None)
        args.update(kwargs)
        for k,v in args.items(): setattr(self, k, v)
        if self.rate == None: self.min, self.max, self.rate = self.nom, self.nom, 1
        if 'min' not in args: self.min = min(max(1, self.nom if self.nom < 12 else self.nom/4), self.max)
        if self.min > self.max or self.rate < 0: raise ValueError(f"Invalid parameter {self}")
        if not self.rate:
            if self.nom < self.min: self.nom = self.min
            elif self.nom > self.max: self.nom = self.max
        #print(f"SIZE {kwargs} -> {self}")


    def clone(self, **kwargs):
        return Size(**{k:kwargs.get(k,self[k]) for k in Size.__ports__})


    def __getitem__(self, k):
        return getattr(self,k)

    def dir(self, free):
        if self.size < self.min - _err: return 1 # We must grow because we are below our minimum
        if self.size > self.max + _err: return -1 # We must shrink because we are above our maximum
        shrink = self.size > self.min + _err # We can shrink if we are above our minimum
        grow = self.size < self.max - _err # We can grow if we are below our maximum
        if free > _err: return int(grow) # There is free space so grow if we can
        if free < -_err: return -int(shrink) # We are overflowing so shrink if we can
        return 2 + int(shrink) + 2*int(grow) # We are at the target size so return an indication of which way we can go
    

    def drate(self, dir, fixed_rate=0):
        if not self.rate: return 0 # No need to go anywhere
        if dir == 0: return 0
        if dir == -1: return -1.0/self.rate
        if dir == 1:  return self.rate
        if not fixed_rate: return 0
        if fixed_rate < 0:
            return self.rate if dir >= 4 else 0
        else:
            return -1.0/self.rate if dir%2 else 0
    
    
    def t_hit(self, rate):
        # TODO: Optimize by not stopping at minimum for positive rates, or maxima for negative rates
        if rate < 0:
            target = self.min if self.size < self.max+_err else self.max
        else:
            target = self.max if self.size > self.min-_err else self.min
        return (target-self.size)/rate


    def ideal(self):
        self.size = min(max(self.min, self.nom), self.max)
        return self


    def __repr__(self):
        s = []
        if self.nom != 0: s.append(f'nom={self.nom}')
        if True: s.append(f'min={self.min}')
        if self.max != 1e10: s.append(f'max={self.max}')
        if self.rate != 1: s.append(f'rate={self.rate}')
        return 'Size(' + ', '.join(s) + f')[{self.size}]'
        

    @staticmethod
    def resize(cells, size, wrap=False, elide=None):
        if size == 0: return [[c.ideal() for c in cells]]
        rows = _resize_wrap(cells, size) if wrap else [_resize_no_wrap(cells, size)]
        return [_elide(row, size, elide) for row in rows] if elide else rows



def _elide(cells, size, elide):
    ''' If sizes overflows then replace the middle portion with the Size returned from elide()'''
    zeros = any(not x.size and x.min for x in cells)
    #print(f"_elide {sum([c.size for c in cells])} -> {size} zeros?:{zeros}", ''.join(f'\n  {c} {c.user}' for c in cells))
    if sum([c.size for c in cells]) <= size and not zeros: return cells
    total = cells[0].size
    r = [1, len(cells)]
    ri = 1
    while r[0] < r[1] and total + cells[r[ri] + 1 - ri*2].size < size:
        r[ri] += 1 - ri*2
        total += cells[r[ri]].size
        ri = (ri+1)%2
    if r[0] != r[1]:
        cells = _resize_no_wrap(cells[:r[0]] + [elide(cells[r[0]:r[1]])] + cells[r[1]:], size)
        print(f"ELIDED {cells}")
    #print(f"    {r} {total}", ''.join(f'\n  {c} {c.user}' for c in cells))
    total = sum([c.size for c in cells])
    if total > size: cells[0].size += size - total
    assert(cells[0].size > 0), f"The elided cell is to big. {cells}"
    return cells



def _round(cells, size):
    if not size: return cells
    sizes = [modf(c.size) for c in cells]
    s = [[i, int(s[1]), s[0], cells[i]] for i,s in enumerate(sizes)]
    sw = sum([x[1] for x in s])
    s = sorted(s, key=lambda x: x[2], reverse=True)
    i = 0
    while sw + i != size and i < len(s) and s[i][2]:
        s[i][1] += 1
        i += 1
    for _,s,_,cell in sorted(s, key=lambda x:x[0]): cell.size = s
    return cells

    
def _resize_wrap(cells, size):
    ci = 0
    rows = []
    while cells:
        #print(f"_resize_wrap size:{size}", ''.join(f'\n  {c} {c.user}' for c in cells))
        _resize_no_wrap(cells, size)
        if len(rows): _resize_no_wrap(rows[-1], size)
        rows.append([cells.pop(0)])
        row_size = rows[-1][0].size
        while cells:
            row_size += cells[0].size
            if row_size > size: break
            rows[-1].append(cells.pop(0))
    return rows


def _resize_no_wrap(cells, size):
    for c in cells: c.size = c.nom
    while True:
        sizes = [c.size for c in cells]   # The current actual sizes of each cell
        free = size - sum(sizes) # The amount of free space to fill
        #print(f"_resize_no_wrap {sizes}/{size}  free:{free}")
        dirs = [c.dir(free) for c in cells] # <=1 we must change,  >1 We might be able to change if needed
        if not any(dirs): return _round(cells, size) # Everyone is static
        rates = [c.drate(d) if d <= 1 else 0 for c,d in zip(cells, dirs)] # At what rate MUST each cell grow or shrink?
        fixed_rate = sum(rates) # The rate at wich the TOTAL size MUST change
        vrates = [c.drate(d, fixed_rate) if d > 1 else 0 for c,d in zip(cells,dirs)] # When we are at size, but we MUST change, then these vrates will compensate for fixed_rate
        variable_rate = sum(vrates) # The rate at which the total moves due to the vrate cells
        if not variable_rate and not fixed_rate: return _round(cells, size) # reached size and everyone inbounds
        if variable_rate: # variable cells are compensating for fixed_rate cells
            rates = [r*fixed_rate/-variable_rate if r else rates[i] for i,r in enumerate(vrates)] # variable rates step in to offset the fixed rate
        total_rate = sum(rates) # This is the rate at wich the total size will change
        ts = [c.t_hit(r) for c,r in zip(cells,rates) if r] # The time when the cell hits a critical junction
        assert(ts)
        tt = 0
        if free and total_rate: # Only fixed rates
            tt = free / total_rate # The amount of time to close the 'free' gap
            if tt >= 0: ts.append(tt)
        t = min(ts)
        assert(t)
        for c,r in zip(cells, rates): c.size += r*t
    



from print_ext.hr import HR
from print_ext.line import SMark as SM
from print_ext.flex import Flex
from print_ext.text import Text
from .printer_test import _printer
from .testutil import flat, printer

def test_hr_hello_pretty():
    o, p = _printer(width=10, ascii=True)
    p.hr('hi')
    assert(o.getvalue() == '--[ hi ]--\n')


def test_hr_blank():
    assert(flat(HR(width_max=10, ascii=True)) == ['----------'])
    assert(flat(HR(border=('t:abcde'))) == ['acd'])
    assert(HR().width == 3)
    assert(HR(width_max=5).width == 5)
    assert(flat(HR('','',width_max=4,ascii=True)) == ['----'])


def test_hr_just_h():
    o,p = _printer(width=10, ascii=True)
    p.hr('hi', just='<')
    print(o.getvalue())
    assert(o.getvalue() == '-[ hi ]---\n')
    h = HR('hi', just='|', ascii=True, width_max=10, style='y')
    assert([f.styled() for f in h.flatten()][0] == ('--[ hi ]--', [SM('y',0,10)]))
    assert([f.styled() for f in h.flatten(w=9)][0] == ('-[ hi ]--', [SM('y',0,9)]))
    h['just'] = ':'
    assert([f.styled() for f in h.flatten(w=9)][0] == ('--[ hi ]-', [SM('y',0,9)]))
    print('\n\n\n\n')
    h['just'] = '>'
    assert([f.styled() for f in h.flatten()][0] == ('---[ hi ]-', [SM('y',0,10)]))



def test_hr_just_v():
    h = HR('a\vbbb\vc\v', ascii=True, width_max=11)
    assert([f.styled()[0] for f in h.flatten()] == ['  [ a   ]  ', '  [ bbb ]  ','--[ c   ]--','  [     ]  '])
    h['just'] = '_'
    assert([f.styled()[0] for f in h.flatten()] == ['  [ a   ]  ', '  [ bbb ]  ','  [ c   ]  ','--[     ]--'])
    h['just'] = '^'
    assert([f.styled()[0] for f in h.flatten()] == ['--[ a   ]--', '  [ bbb ]  ','  [ c   ]  ','  [     ]  '])
    h['just'] = '-'
    h['ascii'] = False
    assert([f.styled()[0] for f in h.flatten()] == ['  │ a   │  ', '──┤ bbb ├──','  │ c   │  ','  │     │  '])
    h['just'] = '<'
    assert([f.styled()[0] for f in h.flatten()] == [' │ a   │   ', ' │ bbb │   ','─┤ c   ├───',' │     │   '])



def test_hr_small():
    h = HR('hello world', ascii=True)
    assert([f.styled()[0] for f in h.flatten()] == ['-[ hello world ]-'])
    assert([f.styled()[0] for f in h.flatten(w=11)] == ['-[ he~ld ]-'])
    assert([f.styled()[0] for f in h.flatten(w=7)] == ['-[ ~ ]-'])
    assert([f.styled()[0] for f in h.flatten(w=6)] == ['hel~ld'])
    h['wrap']=True
    assert([f.styled()[0] for f in h.flatten(w=11)] == [' [ hello ] ', '-[ \\  wo ]-', ' [ \\ rld ] '])



def test_hr_clone():
    h = HR('bob', ascii=False)
    print(h['width_max'])
    assert([str(x) for x in h.flatten()] == ['─┤ bob ├─'])
    assert([str(x) for x in Text(h,h).flatten()] == ['─┤ bob ├──┤ bob ├─'])


def test_hr_infinity():
    o,p = printer(ascii=True)
    p.hr()
    assert(o.getvalue() == '---\n')
